
/* Common tape pragmas, defines, globals and stuff */

ushort_t  inb(int port);
#pragma aux inb = \
          0x32 0xe4          /* xor ah,ah */\
          0xec               /* in  al,dx */\
          parm nomemory [dx] \
          modify exact nomemory [ax];

void outb(int port, ushort_t val);
#pragma aux outb = \
          0xee             /*  out dx,al   */\
          parm nomemory [dx] [ax] \
          modify exact nomemory [];

void	io_safe(void);
#pragma	aux	io_safe = \
			0xeb 0x00	/*	jmp .	*/	\
			0xeb 0x00	/*	jmp .	*/	\
			parm nomemory []      		\
			modify exact nomemory [];

#define JMPNXT	io_safe();

unsigned disable(void);
#pragma aux disable = \
			0x9c				/*	pushf	*/ \
			0xfa				/*	cli		*/ \
			0x58				/*	pop ax  */ \
			parm nomemory	[]				   \
			modify exact nomemory [ AX ];

unsigned enable(void);
#pragma	aux	enable = \
			0x9c				/*	pushf	*/ \
			0xfb				/*	sti		*/ \
			0x58				/*	pop ax  */ \
			parm nomemory []				   \
			modify exact nomemory [ AX ];

#define BLOCKSIZE	 512
#define	TREAD_STAT	0xc0
#define	TREWIND		0x21
#define TERASE		0x22
#define	TRETENSION	0x24
#define TSELECT		0x01
#define TWRITE		0x40
#define TWRITEFM	0x60
#define TREAD		0x80
#define TREADFM		0xa0

#define TSELFTST1   0x0c2
#define TSELFTST2	0x0ca

/*
 * Error status codes.  These are determined from bytes read by read_status.
 */

#define OK                  0
#define NO_CART 			1
#define NO_DRIVE			2
#define WRT_PROTECT			3
#define EOM					4
#define RD_WRT_ABORT		5
#define RD_BBXFER			6
#define RD_FBXFER			7
#define RD_NODATA			8
#define RD_EOM				9
#define RD_FILEMARK			10
#define ILLEGAL				11
#define	POWER_RST			12
#define	MARGINAL			13
#define	BOM					14
#define NOTHING_IN_BUFFER	15
#define ILLEGAL_PAR			16

/* -------DMA values --------------------------------------------------------*/
#define DMA_PORT		0x0000

#define DMAI	0x44		/* io_inut mode 		*/
#define DMAO	0x48		/* io_outut mode		*/
#define DMAE	0x00		/* mask enable  	*/
#define DMAD	0x04		/* mask disable 	*/
#define DMASTAT	0x08		/* status register  */
#define DMACBF	0x0C		/* clear byte 0xff 	*/
#define DMAMBR	0x0A		/* mask bit register*/
#define DMAMOD	0x0B		/* mode				*/
	
static short dmapgaddr [ ] = { 0x87, 0x83, 0x81, 0x82 };

#define BRK_IGNORE  /* turn off breaks */
#define BRK_ALLOW   /* allow breaks */

/* globals */

unsigned CTRL_PORT, STATUS_PORT, CMD_PORT, DATA_PORT;
ushort_t S_RDY, S_EXC, ONL, RESET, REQ;

char online;
int last_cmd;
char srb[6];

#define MAX_SEGS 100
unsigned sel[MAX_SEGS];

/* defaults */
unsigned io_port     = 0x300;
int PS2              = 0;
char unit_type       = 'a';
unsigned dma_channel = 3;

/*************** some prototypes **********************/

void pausetape(int);

/*
 * base io functions, segregated to provide similiar timing delays to 'old'
 * qnx drivers.
 */

ushort_t io_in(ushort_t ioaddr)
{
ushort_t inval;

    inval = inb(ioaddr)&0xff;
#ifdef DEBUG
    fprintf(stdout,"In %x : %x\n",ioaddr,inval);
#endif
    return inval;
}

void io_out(ushort_t ioaddr, ushort_t val)
{
#ifdef DEBUG
    fprintf(stdout,"Out %x : %x\n",ioaddr, val & 0xff);
#endif
    outb(ioaddr, val & 0xff);
}

