#ifdef __USAGE
%C - QIC-02 Tape manager

%C -t Unit_type [-a io_port] [-d DMA Channel] [-V] &
Options:
 -t         Type of Unit: Choices are: Archive, PS2, External, Internal
 -a         I/O port address of tape unit. Entered value is hexadecimal.
            Default is 0x300.
 -d         DMA Channel of Tape unit. Entered value is hexadecimal. 
            Default is 3.
 -b         Number of 32K blocks to use as buffer. Default = 1.
 -V         Run verbose. All debugging info goes to stderr.

e.g. %C -t External -a 0x200 -d 3 -b 30 &
#endif

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <sys/psinfo.h>
#include <sys/osinfo.h>
#include <sys/sys_msg.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <sys/prfx.h>
#include <sys/sched.h>
#include <sys/io_msg.h>
#include <sys/fd.h>
#include <time.h>
#include <sys/seginfo.h>

#include "tapedef.h"

#define		PFLAGS	(_PPF_SERVER|_PPF_PRIORITY_REC)

struct	_sysmsg_version_reply	version_stamp = {
		"Tapedev",
		__DATE__,
		410,
		'X',
		0
};

union {
	msg_t						type;
	msg_t						status;
	struct _sysmsg_hdr			sysmsg;
	struct _sysmsg_hdr_reply	version_reply;
	struct _io_open				open;
	struct _io_open_reply		open_reply;
	struct _io_write			write;
	struct _io_write_reply		write_reply;
	struct _io_read				read;
	struct _io_read_reply		read_reply;
	struct _io_fstat			fstat;
	struct _io_fstat_reply		fstat_reply;
	struct _io_utime			utime;
	struct _io_utime_reply		utime_reply;
	struct _io_chmod			chmod;
	struct _io_chmod_reply		chmod_reply;
	struct _io_chown			chown;
	struct _io_chown_reply		chown_reply;
	struct _io_lseek			lseek;
	struct _io_lseek_reply		lseek_reply;
	struct _io_close			close;
	struct _io_close_reply		close_reply;
	struct _io_dup				dup;
	struct _io_dup_reply		dup_reply;
} msg;

struct _io_fstat_reply fstat_reply;

struct ocb {
	mode_t		mode;
	int			count;
	off_t		offset;
};

struct _osinfo osdata;
struct _psinfo psdata;
struct _psinfo3 psdata3;

struct _fd_ctrl	*fd_ctrl;
time_t __far *timep;

unsigned endmem = 65535;

mode_t perm_mode[4] = { S_IREAD, S_IWRITE, S_IREAD|S_IWRITE };

int verbose = 0;       /* verbosity level */
int opencount = 0;     /* number of processes that have the device open */

int num_segs = 1;
int curr_seg = 0;

char reading, writing;
char done_a_buffer;    /* set to 0 on each open, set to 1 when some data is written */

int segsizes[MAX_SEGS];

void main(int argc, char *argv[]) {
	struct _mxfer_entry mx_entry[2];
	int devno;
	pid_t pid;
	msg_t type;
	int nbytes;
	mode_t mode;
	struct ocb *ocb;
	int size;
    int opt;
    char tapetype;

	while ((opt = getopt(argc,argv,"b:t:a:d:V")) != -1) {
		switch(opt) {
			case 't': tapetype = optarg[0];
					  break;
			case 'a': io_port = strtol( optarg,NULL,16);
					  break;
			case 'b': num_segs = atoi( optarg );
                      if ( num_segs > MAX_SEGS ) {
                         fprintf(stderr,"Max of %d segments allowed.\n",MAX_SEGS);
                         exit(-1);
                         }
					  break;
			case 'd': dma_channel = strtol( optarg,NULL,16);
                      break;
            case 'V': verbose++;
                      break;
			default:  break;
		}
	}

    switch ( tapetype ) {
        case 'A':
        case 'a': /* archive */
                  PS2 = 0;
	              unit_type = 'a';
                  CTRL_PORT = STATUS_PORT = io_port + 1;
		          CMD_PORT = DATA_PORT = io_port;
		          S_RDY = 0x40; S_EXC = 0x20; ONL = 0x00; RESET = 0x80; REQ = 0x40;
                  break;
        case 'E':
        case 'e': /* external */
                  PS2 = 0;
	              unit_type = 'e';
		          CTRL_PORT = STATUS_PORT = io_port;
		          CMD_PORT  = DATA_PORT = io_port + 1;
		          S_RDY = 0x01; S_EXC = 0x02; ONL = 0x01; RESET = 0x02; REQ = 0x04;
                  break;
        case 'I':
        case 'i': /* internal */
                  PS2 = 0;
	              unit_type = 'i';
		          CTRL_PORT = STATUS_PORT = io_port;
		          CMD_PORT  = DATA_PORT = io_port + 1;
		          S_RDY = 0x01; S_EXC = 0x02; ONL = 0x01; RESET = 0x02; REQ = 0x04;
                  break;
        case 'P':
        case 'p': /* PS/2 */
                  PS2 = 1;
	              unit_type = 'e';
		          CTRL_PORT = STATUS_PORT = io_port;
		          CMD_PORT  = DATA_PORT = io_port + 1;
		          S_RDY = 0x01; S_EXC = 0x02; ONL = 0x01; RESET = 0x02; REQ = 0x04;
                  break;
        default : /* none specified */
                  fprintf(stderr,"You must specify a tape unit type (-t)\n");
                  exit(-1);
		}

    if ( io_in(STATUS_PORT) == 0xff ) {
       fprintf(stderr,"No controller found at port 0x%x\n",io_port);
       exit(-1);
       }

	if(qnx_osinfo(0, &osdata) == -1) {
		perror("osinfo");
		exit(EXIT_FAILURE);
	}
	timep = MK_FP(osdata.timesel, offsetof(struct _timesel, seconds));

	if(qnx_psinfo(PROC_PID, 0, &psdata, 0, NULL) == -1) {
		perror("psinfo");
		exit(EXIT_FAILURE);
	}

	if((devno = qnx_device_attach()) == -1) {
		perror("device_attach");
		exit(EXIT_FAILURE);
	}

	fstat_reply.status = EOK;
	fstat_reply.stat.st_rdev = (dev_t) ((getnid() << 16L) + (devno << 10L)) + 1;
	fstat_reply.stat.st_mode = S_IFREG|S_IREAD|S_IWRITE;
	fstat_reply.stat.st_size = 65535;
	fstat_reply.stat.st_nlink = 1;
	fstat_reply.stat.st_ftime =
	fstat_reply.stat.st_mtime =
	fstat_reply.stat.st_atime =
	fstat_reply.stat.st_ctime = *timep;

	if(qnx_pflags(PFLAGS, PFLAGS, NULL, NULL)) {
		perror("pflags");
		exit(EXIT_FAILURE);
	}

	if(!(fd_ctrl = __init_fd(psdata.pid))) {
		perror("__init_fd");
		exit(EXIT_FAILURE);
	}

	if(qnx_prefix_attach("/dev/qic02", NULL, 0)) {
		perror("prefix_attach");
		exit(EXIT_FAILURE);
	}

/*
	qnx_scheduler(0, 0, SCHED_RR, 22, 1); 
*/

/*
	(void)signal(SIGHUP, SIG_IGN);
    if(signal(SIGINT,  SIG_IGN) == SIG_IGN) {
		close(0);
		close(1);
		close(2);
	}
*/

    get_tape_buffer();

	for(;;) {
		pid = Receive(0, &msg, sizeof(msg));
		type = msg.type;
		msg.status = ENOSYS;
		nbytes = sizeof(msg.status);
		switch(type) {

		case _SYSMSG:
			switch(msg.sysmsg.subtype) {
			case _SYSMSG_SUBTYPE_VERSION:
				msg.version_reply.status = EOK;
				_setmx(&mx_entry[0], &msg, sizeof(msg.version_reply));
				_setmx(&mx_entry[1], &version_stamp, sizeof(version_stamp));
				Replymx(pid, 2, &mx_entry);
				nbytes = 0;
				break;
			}
			break;

		case _IO_HANDLE:
			switch(mode = msg.open.oflag) {
			case _IO_HNDL_INFO:
				goto open2;
			case _IO_HNDL_CHANGE:
				__get_pid_info(pid, &psdata3, fd_ctrl);
				if(psdata3.euid != 0) {
	 				msg.status = EPERM;
 					break;
				}
				goto open2;
			case _IO_UTIME:
				mode |= S_IWRITE;
				goto open1;
			}
			break;

		case _IO_OPEN:
            if ( verbose )
               fprintf(stderr,"Request to open\n");

			mode = perm_mode[msg.open.oflag & O_ACCMODE];
open1:
			__get_pid_info(pid, &psdata3, fd_ctrl);
			mode &= (mode & 077) | /* re-use the lower 6 bits for handle */
				(psdata3.euid == 0) ?
					S_IRWXU :
				(psdata3.euid == fstat_reply.stat.st_uid) ?
					(fstat_reply.stat.st_mode & S_IRWXU) :
				(psdata3.egid == fstat_reply.stat.st_gid) ?
					(fstat_reply.stat.st_mode & S_IRWXG) << 3 :
					(fstat_reply.stat.st_mode & S_IRWXO) << 6;
			if((mode & S_IRWXU) == 0) {
 				msg.status = EPERM;
 				break;
			}
open2:
			if(msg.open.path[0] != '\0') {
 				msg.status = ENOENT;
 				break;
			}
            if ( opencount ) {  /* someone already has it open */
               if ( verbose )
                  fprintf(stderr,"device busy\n");
               msg.status = EBUSY;
               break;
            }
			if(!(ocb = (struct ocb *)malloc(sizeof(struct ocb)))) {
				msg.status = ENOMEM;
				break;
			}
			memset(ocb, 0, sizeof(struct ocb));
			ocb->mode = mode;
			ocb->count++;
			if(qnx_fd_attach(pid, msg.open.fd,
						0, 0, 0, 0, (unsigned)ocb) == -1) {
				free(ocb);
				msg.status = errno;
				break;
			}
			msg.status = EOK;						
            opencount++;
            done_a_buffer = 0;
            reading = writing = 0;
			break;

		case _IO_DUP:
            if ( opencount ) {  /* someone already has it open */
               if ( verbose )
                  fprintf(stderr,"Device busy\n");
               msg.status = EBUSY;
               break;
            }
			ocb = (struct ocb *)__get_fd(
					msg.dup.src_pid ? msg.dup.src_pid : pid,
					msg.dup.src_fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if(qnx_fd_attach(pid, msg.dup.dst_fd,
						0, 0, 0, 0, (unsigned)ocb) == -1) {
				msg.status = errno;
				break;
			}
			ocb->count++;
			msg.status = EOK;						
			break;

		case _IO_CLOSE:
            if ( verbose )
               fprintf(stderr,"Request to close\n");

			ocb = (struct ocb *)__get_fd(pid, msg.close.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if(--ocb->count == 0) {
				free(ocb);
			}
            opencount = 0;

            if ( writing && !done_a_buffer )
               last_cmd = TWRITE;

            switch ( last_cmd ) {
               case TWRITE: /* flush any pending buffers */
                            tape_write(0,curr_seg);
                            /* write a file mark */
	                        online = ONL;
                            if ( verbose )
                               fprintf(stderr,"Writing file mark\n");
	                        tape_simple_cmd(TWRITEFM);
                            if ( verbose )
                               fprintf(stderr,"Reset after write.\n");
	                        tape_reset();
                            break;
               case TREAD:  /* possible abortion of read ... */
                            if ( verbose )
                               fprintf(stderr,"Reset after read.\n");
	                        tape_reset();
                            break;
	           }
			msg.status = EOK;
			break;

		case _IO_STAT:
			if(msg.open.path[0] != '\0') {
 				msg.status = ENOENT;
 				break;
			}
			Reply(pid, &fstat_reply, sizeof(fstat_reply));
			nbytes = 0;
			break;

		case _IO_FSTAT:
			ocb = (struct ocb *)__get_fd(pid, msg.fstat.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			Reply(pid, &fstat_reply, sizeof(fstat_reply));
			nbytes = 0;
			break;

		case _IO_CHMOD:
			ocb = (struct ocb *)__get_fd(pid, msg.chmod.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if((ocb->mode & 077) != _IO_HNDL_CHANGE) {
				msg.status = EBADF;
				break;
			}
			fstat_reply.stat.st_mode = msg.chmod.mode;
			msg.chmod_reply.status = EOK;
			nbytes = sizeof(msg.chmod_reply);
			break;
			
		case _IO_CHOWN:
			ocb = (struct ocb *)__get_fd(pid, msg.chown.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if((ocb->mode & 077) != _IO_HNDL_CHANGE) {
				msg.status = EBADF;
				break;
			}
			fstat_reply.stat.st_uid = msg.chown.uid;
			fstat_reply.stat.st_gid = msg.chown.gid;
			msg.chown_reply.status = EOK;
			nbytes = sizeof(msg.chown_reply);
			break;

		case _IO_UTIME:
			ocb = (struct ocb *)__get_fd(pid, msg.utime.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if((ocb->mode & S_IWRITE) == 0) {
				msg.status = EBADF;
				break;
			}
			if(msg.utime.cur_flag) {
				fstat_reply.stat.st_mtime =
				fstat_reply.stat.st_atime = *timep;
			} else {
				fstat_reply.stat.st_mtime = msg.utime.modtime;
				fstat_reply.stat.st_atime = msg.utime.actime;
			}
			msg.utime_reply.status = EOK;
			nbytes = sizeof(msg.utime_reply);
			break;

		case _IO_LSEEK:
			msg.status = EOK;
			break;

		case _IO_READ:
			ocb = (struct ocb *)__get_fd(pid, msg.read.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if((ocb->mode & S_IREAD) == 0) {
				msg.status = EBADF;
				break;
			}
			size = msg.read.nbytes;
            reading = 1;

			_setmx(&mx_entry[1], MK_FP(sel[0], 0), size);
            tape_read(size);

			msg.read_reply.status = EOK;
			msg.read_reply.nbytes = size;
			msg.read_reply.zero   = 0;
			_setmx(&mx_entry[0], &msg,
				sizeof(struct _io_read_reply) - sizeof(msg.read_reply.data));
			Replymx(pid, 2, &mx_entry);
			nbytes = 0;
			fstat_reply.stat.st_atime = *timep;
			break;

		case _IO_WRITE:
			ocb = (struct ocb *)__get_fd(pid, msg.write.fd, fd_ctrl);
			if(ocb == (struct ocb *)0 || ocb == (struct ocb *)-1) {
				msg.status = EBADF;
				break;
			}
			if((ocb->mode & S_IWRITE) == 0) {
				msg.status = EBADF;
				break;
			}

			size = msg.write.nbytes;
            writing = 1;

			_setmx(&mx_entry[0], MK_FP(sel[curr_seg], 0), size);
            segsizes[curr_seg] = size;

			msg.write_reply.nbytes = Readmsgmx(pid,
					sizeof(struct _io_write) - sizeof(msg.write.data),
					1, &mx_entry);

            curr_seg++;
            if ( curr_seg == num_segs ) {
               tape_write(0,curr_seg);  /* do them all */
               curr_seg = 0;
               }

			if(msg.write_reply.nbytes == -1) {
				msg.write_reply.status = errno;
				nbytes = sizeof(msg.write_reply.status);
			} else {
				msg.write_reply.status = EOK;
				nbytes = sizeof(msg.write_reply);
			}
			break;
		}
		if(nbytes) {
			Reply(pid, &msg, nbytes);
		}
	}
}

/***************************************************************************/

tape_reset() {
	int i;

	online = 0;
	io_out(CTRL_PORT, RESET);			/* Reset line TRUE */
	pausetape(25);							/* Wait at least 25 usec */
	wait_till_not_rdy();				/* Wait for NOT ready */
	while(isexc() == 0)					/* Wait for exception */
		JMPNXT;
	io_out(CTRL_PORT, online);			/* Reset line FALSE */

	if((i = tape_read_status()) != POWER_RST)	/* Clear the status bytes */
        if ( verbose )
		   report_status(i);				/* Print any errors */
	}

tape_simple_cmd(cmd)
unsigned char cmd;
{
	int status = 0;

	wait_till_rdy();
	io_out(CTRL_PORT, online);
	io_out(CMD_PORT, last_cmd = cmd);
	io_out(CTRL_PORT, REQ | online);
	wait_till_rdy();
	io_out(CTRL_PORT, online);
	pausetape(100);					/* Wait at least 100 usec */

	for(;;) {
		if(isrdy())
			break;
		if(isexc()) {
			status = tape_read_status();
            if ( verbose )
     			report_status(status);
			break;
		}
	}

	return(status);
}

wait_till_rdy() {

	while(isrdy() == 0)
		JMPNXT;
	}

wait_till_not_rdy() {
	while( isrdy() )
		JMPNXT;
	}

isrdy() {
	return((io_in(STATUS_PORT) & S_RDY) == 0);		/* Bit 0 when ready */
	}

isexc() {
	return((io_in(STATUS_PORT) & S_EXC) == 0);		/* Bit 0 when execption */
	}

/*
 * pausetape for at least the passed number of usec. We actually pausetape
 * a fair bit longer to play it safe.
 */

void pausetape(int usecs)
{
int ios;

usecs *=2;
ios = usecs + usecs/4 + ((usecs & 3) != 0);
    while (ios--) {
        inb(0);
    }
}

tape_read(bytes)
int bytes;
	{
	unsigned status = OK, tc;
    int count;
	
	tape_rw_cmd(TREAD,0);
    if ( verbose )
       fprintf(stderr,"Trying to read %d bytes\n",bytes);

for ( count =0; count < bytes; count+=BLOCKSIZE ) {
    setup_dma(0, sel[0], count); 

	while((tc = istc()) == 0  &&  isexc() == 0) {
       if ( verbose )
          fprintf(stderr,"spinning in read... tc=%d exc=%d\n",tc,isexc() );
       }

	while(isrdy() == 0  &&  isexc() == 0) {
		Yield();	/* Without this DMA xfer screwed up! */
        }
	disable_dma();
	if(isexc()) {
		status = tape_read_status();
        if ( verbose )
			report_status(status);
		tape_rw_cmd(TREAD,0);
		}
    }

	return(status);
	}

tape_write(startseg, endseg)
unsigned startseg, endseg;
	{
	unsigned status = OK, tc;
    int count;
	int segcount;
    char firsttime = 1;

  done_a_buffer = 1;
  for ( segcount = startseg; segcount < endseg; segcount++ ) {
    if ( verbose )
       fprintf(stderr,"Writing %d bytes from seg %d\n",segsizes[segcount],segcount);
    for ( count =0; count < segsizes[segcount]; count+=BLOCKSIZE ) {

    	tape_rw_cmd(TWRITE,firsttime);
        firsttime = 0;
	    disable_dma();

        setup_dma(1, sel[segcount], count); 

	    while((tc = istc()) == 0  &&  isexc() == 0) {
          if ( verbose )
             fprintf(stderr,"spinning in write..."); 
          }

	    while(isrdy() == 0  &&  isexc() == 0) {
		   Yield();	/* Without this DMA xfer screwed up! */
           }
	    disable_dma();

	    if(isexc()) {
		   status = tape_read_status();
           if ( verbose )
			  report_status(status);
		   tape_rw_cmd(TWRITE,0);
		   }
        }
    }
}

tape_rw_cmd(cmd,forceit)
char cmd;
char forceit;
	{
	int status, i;

	if((cmd == last_cmd) && !forceit)
		return;

	disable_dma();
	while(isrdy() == 0) 
		if(isexc()) {
			status = tape_read_status();
            if ( verbose )
			   report_status(status);
			}

    last_cmd = cmd;
	io_out(CMD_PORT, last_cmd);
	online = ONL;
	io_out(CTRL_PORT, online);
	io_out(CTRL_PORT, REQ | online);
	wait_till_rdy();
	io_out(CTRL_PORT, online);
	pausetape( 100 );

	for(i = 0 ; i < 3000 ; ++i) {	/* Wait for not rdy or a couple 100 usec */
		if(isrdy() == 0)
			break;
		}

	while(isrdy() == 0  &&  isexc() == 0) {
		Yield();
        }
	}

tape_read_status() {
	int i;

	io_out(CMD_PORT, last_cmd = TREAD_STAT);/* Put command on the bus */
	io_out(CTRL_PORT, REQ | online);		/* Set request and online */
	wait_till_rdy();
	io_out(CTRL_PORT, online);				/* Reset request but keep online */
	/*
	 * At this point the ready line will go from ready to not-ready
	 * It can take 100 usec before it goes not ready.
	 */
	pausetape(100);								/* Wait at least 100 usec */
	for(i = 0 ; i < 6 ; ++i) {
		wait_till_rdy();
		srb[i] = io_in(DATA_PORT);			/* Get the status */
		io_out(CTRL_PORT, REQ + online);	/* Set request */
		wait_till_not_rdy();
		io_out(CTRL_PORT, online);			/* Reset request */
		}

	wait_till_rdy();
	return(map_status());
	}

setup_dma(write, seg, offset)
unsigned	write;
unsigned	seg;
unsigned    offset;
	{
    struct _seginfo pbuf;
union {
    paddr_t    addr;
    unsigned char bytes[4];
} physaddr;
	
    qnx_segment_info(0,0,seg,&pbuf);
    physaddr.addr = pbuf.addr + offset;
    if ( verbose ) {
/*
       fprintf(stderr,"setting up dma ... %lx, %x  ",physaddr.addr, offset);
       fprintf(stderr,"[0]=%x ",physaddr.bytes[0]);
       fprintf(stderr,"[1]=%x ",physaddr.bytes[1]);
       fprintf(stderr,"[2]=%x ",physaddr.bytes[2]);
       fprintf(stderr,"[3]=%x\n",physaddr.bytes[3]);
*/
       }

    disable();

	io_out(DMA_PORT + DMAMBR, DMAD | dma_channel);	/* disable dma channel x on 8237	*/
	JMPNXT;
	io_out(DMA_PORT + DMACBF, 0);				/*	initialize first_last ff */
	JMPNXT;
	io_out(DMA_PORT + DMAMOD, (write ? DMAO : DMAI) + dma_channel);
	JMPNXT; 
    io_out(DMA_PORT + dma_channel*2, physaddr.bytes[0]);
	JMPNXT; 
    io_out(DMA_PORT + dma_channel*2, physaddr.bytes[1]);
	JMPNXT; 
	io_out( dmapgaddr[dma_channel], physaddr.bytes[2] );
	JMPNXT;
	io_out(DMA_PORT + dma_channel*2 + 1, 0xff);
	JMPNXT;
	io_out(DMA_PORT + dma_channel*2 + 1, 0x01);	/*	byte count is 512		*/
	JMPNXT;

    if ( PS2 ) {
       io_in(DMA_PORT + DMASTAT);
       JMPNXT;
       io_out(DMA_PORT + DMAMBR, dma_channel);
       JMPNXT;
       }

	if(unit_type == 'a')
		io_out(CMD_PORT + 2, 0);
	else if(unit_type == 'e'  ||  unit_type == 'i')
		io_out(CTRL_PORT, dma_channel != 3 || unit_type == 'e' ? 0x89 : 0x91);
	else
		fprintf(stderr,"Tape drive type is not known\n");

    if ( !PS2 ) {
       io_in(DMA_PORT + DMASTAT);
       JMPNXT;
       io_out(DMA_PORT + DMAMBR, dma_channel);
       JMPNXT;
       }

    enable();
	}

disable_dma() {
    if ( PS2 ) {
        io_out(DMA_PORT + DMAMBR, dma_channel | DMAD);  /* disable channel */
        }
    JMPNXT;
	if(unit_type == 'a')
		io_out(CMD_PORT + 3, 0);				/* disable dma */
	else
		io_out(CTRL_PORT, online);				/* disable dma but keep online		*/
    JMPNXT;

    if ( !PS2 ) {
		io_out(DMA_PORT + DMAMBR, dma_channel | DMAD);
        }
	}

istc() {
	/* Check for terminal count bit */
	return(io_in(DMA_PORT + DMASTAT) & (1 << dma_channel));
	}

map_status() {
	unsigned errnum = ((srb[0] << 8) & 0xff00) | (srb[1] & 0xff) | 0x8000;

	if(srb[1] == 0  &&  (srb[0] & 0x80)) {
		if(srb[0] & 0x20)
			errnum = NO_DRIVE;
		else if(srb[0] & 0x40)
			errnum = NO_CART;
		else if(srb[0] & 0x01)
			errnum = RD_FILEMARK;
		else if(srb[0] & 0x08)
			errnum = EOM;
		else if(srb[0] & 0x10)
			errnum = WRT_PROTECT;
		else if(srb[0] & 0x06)
			errnum = (srb[0] & 0x04) ? RD_FBXFER : RD_BBXFER;
		}
		
	if(srb[1] & 0x80) {
		if(srb[1] & 0x01)
			errnum = POWER_RST;
		else if((srb[0] & 0x90) == 0x90  &&  (srb[1] & 0x77) == 0)
			errnum = WRT_PROTECT;
		else if(srb[1] & 0x08)
			errnum = (srb[0] & 0x84) == 0x84 ? RD_WRT_ABORT : BOM;
		else if(srb[1] & 0x20)
			errnum = (srb[0] & 0x08) ? RD_EOM : RD_NODATA;
		else if(srb[1] & 0x40)
			errnum = ILLEGAL;
		else if(srb[1] & 0x10)
			errnum = MARGINAL;
		else if(srb[0] & 0x08)
			errnum = EOM;
		}
    if ( verbose )
       fprintf(stderr, "Mapping status bytes %2.2x %2.2x --> %4.4x\n",
		     			srb[0] & 0xff, srb[1] & 0xff, errnum);
	return(errnum);
	}

report_status(error)
unsigned error;
	{
	register char *p = NULL;
	static char *error_strs[] ={
		"OK",
		"No cartrige in drive",
		"Drive is off or disconnected",
		"Cartridge is write protected",
		"End of tape",
		"Read or write abort has occurred",
		"Read error, bad block transfer",
		"Read error, filler block transfer",
		"No data detected on this section of tape",
		"End of tape on read",
		"Filemark was read",
		"Tape unit reports that an illegal command was attempted",
		"Power on reset completed",
		"Marginal block detected",
		} ;	

	if(error < sizeof(error_strs)/sizeof(char *))
		p = error_strs[error];
	else
		p = error_strs[0];

    if ( verbose )
       fprintf(stderr, "%s (%4.4X).\n", p, error);
	}

get_tape_buffer() {
int i;

    for ( i=0; i < num_segs; i++ ) {
       sel[i] = qnx_segment_alloc_flags(32767,_PMF_DMA_SAFE|_PMF_DATA_RW);
       if ( sel[i] == -1 ) {
          fprintf(stderr,"Unable to allocate segment %d: %s\n",i,strerror(errno));
          exit(-1);
          }
       if ( verbose )
          fprintf(stderr,"Segment[%d]=%x\n",i,sel[i]);
       }
	}

